--***************************************************************
--* File         : TB_WBUART.vhd                                *
--* Created      : time-stamp: <08/04/04 02:24:44 kawasan>      *
--* Modified     : Time-stamp: <08/06/27 00:22:27 kawasan>      *
--*								*
--* Test-Bench                                                  *
--*								*
--* Copyright 2008 Osamu Kawashima. all rights reserved		*
--***************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

use std.textio.all;
use ieee.std_logic_textio.all;

entity TB_WBUART is end;

architecture behavior of TB_WBUART is
    
	component WBUART
	generic (
		DIVIDER_INIT    : std_logic_vector(11 downto 0) := x"034"
	);
	port (
		-- WISHBONE SLAVE interface:
		WB_CLK_I        : in    std_logic;
		WB_RST_I        : in    std_logic;
		WB_ADR_I        : in    std_logic_vector(7 downto 0);
		WB_DAT_I        : in    std_logic_vector(7 downto 0);
		WB_WE_I         : in    std_logic;
		WB_STB_I        : in    std_logic;
		WB_ACK_O        : out   std_logic;
		WB_DAT_O        : out   std_logic_vector(7 downto 0);
		WB_IRQ_O        : out   std_logic;

		-- Output port (non-WISHBONE signals):
		RXD             : in    std_logic;
		TXD             : out   std_logic;

		TXDONE          : out   std_logic;	-- 1pulse is Tx-Done
		TXEMPTY         : out   std_logic;	-- 0/1 = Tx not empty/Tx empty(Tx-Done)
		RXREADY         : out   std_logic;	-- 0/1 = Rx not ready/Rx ready
		RXFERR          : out   std_logic;	-- Rx Flame Error
		RXOERR          : out   std_logic	-- Rx OverRun Error
	);
	end component;

	constant ZERO32         : std_logic_vector(31 downto 0) := (others=>'0');
	constant ONE32          : std_logic_vector(31 downto 0) := (others=>'1');

	constant DIVIDER_INIT   : std_logic_vector(11 downto 0) := x"034";	-- 57600bps

	constant WT1            : Time := 10 ns;
	constant WT_DELAY       : Time := 1 ns;
	constant WT_IDLE        : Time := 100 ns;
	constant CLK_CYCLE      : Time := 21 ns;	-- about 48MHz

	-- Address
	constant C_DATA         : std_logic_vector := x"00";
	constant C_STAT         : std_logic_vector := x"01";
	constant C_DIV_L        : std_logic_vector := x"02";
	constant C_DIV_H        : std_logic_vector := x"03";
	constant C_CTRL         : std_logic_vector := x"04";
	constant C_DATA9        : std_logic_vector := x"05";

	signal WB_CLK_I         : std_logic := '0';
	signal WB_RST_I         : std_logic := '1';
	signal WB_ADR_I         : std_logic_vector(7 downto 0) := (others=>'0');
	signal WB_DAT_I         : std_logic_vector(7 downto 0) := (others=>'0');
	signal WB_WE_I          : std_logic := '0';
	signal WB_STB_I         : std_logic := '0';
	signal WB_ACK_O         : std_logic := '0';
	signal WB_DAT_O         : std_logic_vector(7 downto 0) := (others=>'0');
	signal WB_IRQ_O         : std_logic := '0';
	signal RXD              : std_logic := '0';
	signal TXD              : std_logic := '0';
	signal TXDONE           : std_logic := '0';
	signal TXEMPTY          : std_logic := '0';
	signal RXREADY          : std_logic := '0';
	signal RXFERR           : std_logic := '0';
	signal RXOERR           : std_logic := '0';

	signal rd_data          : std_logic_vector(7 downto 0) := (others=>'0');
	signal rd_dat           : std_logic_vector(7 downto 0) := (others=>'0');

	-- WB Read (8bit)
	procedure rd_ub(
		ADD_I		: in    std_logic_vector(7 downto 0);
		DAT_I		: in    std_logic_vector(7 downto 0);
	signal	ADD_O		: out   std_logic_vector(7 downto 0);
	signal	DAT_O		: out   std_logic_vector(7 downto 0);
	signal	STB		: out   std_logic;
	signal	WE		: out   std_logic
	) is
	begin
		wait until WB_CLK_I'event and WB_CLK_I='1';
		wait for WT_DELAY;
		ADD_O	<= ADD_I;
		STB	<= '1';
		WE	<= '0';
		wait until WB_CLK_I'event and WB_CLK_I='1';
		wait for WT_DELAY;
		DAT_O	<= DAT_I;
		wait for WT_DELAY;
		ADD_O	<= (others=>'0');
		STB	<= '0';
		WE	<= '0';
	end rd_ub;

	-- WB Write (8bit)
	procedure wr_ub(
		ADD_I		: in    std_logic_vector(7 downto 0);
		DAT_I		: in    std_logic_vector(7 downto 0);
	signal	ADD_O		: out   std_logic_vector(7 downto 0);
	signal	DAT_O		: out   std_logic_vector(7 downto 0);
	signal	STB		: out   std_logic;
	signal	WE		: out   std_logic
	) is
	begin
		wait until WB_CLK_I'event and WB_CLK_I='1';
		wait for WT_DELAY;
		ADD_O	<= ADD_I;
		DAT_O	<= DAT_I;
		STB	<= '1';
		WE	<= '1';
		wait until WB_CLK_I'event and WB_CLK_I='1';
		wait for WT_DELAY;
		ADD_O	<= (others=>'0');
		DAT_O	<= (others=>'0');
		STB	<= '0';
		WE	<= '0';
	end wr_ub;

begin
    
	uWBUART : WBUART
	generic map (
		DIVIDER_INIT    => DIVIDER_INIT
	)
	port map (
		-- WISHBONE SLAVE interface:
		WB_CLK_I        => WB_CLK_I,
		WB_RST_I        => WB_RST_I,
		WB_ADR_I        => WB_ADR_I,
		WB_DAT_I        => WB_DAT_I,
		WB_WE_I         => WB_WE_I,
		WB_STB_I        => WB_STB_I,
		WB_ACK_O        => WB_ACK_O,
		WB_DAT_O        => WB_DAT_O,
		WB_IRQ_O        => WB_IRQ_O,

		-- Output port (non-WISHBONE signals):
		RXD             => RXD,
		TXD             => TXD,

		TXDONE          => TXDONE,
		TXEMPTY         => TXEMPTY,
		RXREADY         => RXREADY,
		RXFERR          => RXFERR,
		RXOERR          => RXOERR
	);

	RXD <= TXD;

	process(WB_RST_I,WB_CLK_I)
	begin
		if WB_RST_I='1' then
			rd_dat <= (others=>'0');
		elsif WB_CLK_I'event and WB_CLK_I='1' then
			if WB_STB_I='1' and WB_WE_I='0' then
				rd_dat <= WB_DAT_O;
			else
				rd_dat <= rd_dat;
			end if;
		end if;
	end process;

	process begin
		WB_CLK_I <= '0';
		wait for CLK_CYCLE/2;
		WB_CLK_I <= '1';
		wait for CLK_CYCLE/2;
	end process;

	process
		variable ln         : line;
	begin
		-- Reset
		wait for WT_IDLE+WT_DELAY;
		WB_RST_I <= '0';
		wait for WT_IDLE;

		-- Init(8bit-Data)
		wr_ub(C_DIV_H, x"00", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);
		wr_ub(C_DIV_L, x"34", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- 57600bps
		wr_ub(C_CTRL, x"02", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- IRQ=RxReady
		wr_ub(C_DATA, x"55", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- Txd=0x55

		-- IRQ-Handler
		for iCnt in 1 to 3 loop
			wait until WB_IRQ_O'event and WB_IRQ_O='1';
			wait for WT_IDLE;
			rd_ub(C_STAT, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);
			rd_ub(C_DATA, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);

			write(ln, string'("C_DATA = "));
			hwrite(ln, rd_dat);
			writeline(output, ln);

			rd_ub(C_STAT, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);

			wr_ub(C_DATA, x"89", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- Txd=0x89
		end loop;

		-- Init(9bit-Data)
		wr_ub(C_DIV_H, x"00", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);
		wr_ub(C_DIV_L, x"30", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- 62.5Kbps
		wr_ub(C_CTRL, x"12", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- 9bit, IRQ=RxReady
		wr_ub(C_DATA9, x"00", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);
		wr_ub(C_DATA, x"12", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- Txd=0x012

		-- IRQ-Handler
		for iCnt in 1 to 3 loop
			wait until WB_IRQ_O'event and WB_IRQ_O='1';
			wait for WT_IDLE;
			rd_ub(C_STAT, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);

			write(ln, string'("C_DATA = "));
			rd_ub(C_DATA9, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);
			hwrite(ln, rd_dat);
			rd_ub(C_DATA, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);
			hwrite(ln, rd_dat);
			writeline(output, ln);

			rd_ub(C_STAT, WB_DAT_O, WB_ADR_I, rd_data, WB_STB_I, WB_WE_I);

			wr_ub(C_DATA9, x"01", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);
			wr_ub(C_DATA, x"23", WB_ADR_I, WB_DAT_I, WB_STB_I, WB_WE_I);	-- Txd=0x123
		end loop;

		-- End
		write(ln, string'("--- End of Test-Bench. ---"));
		writeline(output, ln);
		wait;
	end process;

end behavior;
