--***************************************************************
--* File         : WBUART.vhd                                   *
--* Created      : time-stamp: <05/11/29 21:32:29 yaegashi>     *
--* Modified     : Time-stamp: <08/06/27 00:08:15 kawasan>      *
--*                                                             *
--* WISHBONE UART Controller                                    *
--*                                                             *
--* Serial-Tx/Rx                                                *
--*                                                             *
--* WB_CLK_I / (BPS   * 16) = DIVIDER                           *
--* 48e6     / (57600 * 16) = 52 = 0x34                         *
--* DIVIDER_INIT = x"034"                                       *
--*                                                             *
--* Copyright 2007-2008 Osamu Kawashima. all rights reserved    *
--***************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WBUART is
	generic (
		DIVIDER_INIT    : std_logic_vector(11 downto 0) := x"034"
	);
	port (
		-- WISHBONE SLAVE interface:
		WB_CLK_I        : in    std_logic;
		WB_RST_I        : in    std_logic;
		WB_ADR_I        : in    std_logic_vector(7 downto 0);
		WB_DAT_I        : in    std_logic_vector(7 downto 0);
		WB_WE_I         : in    std_logic;
		WB_STB_I        : in    std_logic;
		WB_ACK_O        : out   std_logic;
		WB_DAT_O        : out   std_logic_vector(7 downto 0);
		WB_IRQ_O        : out   std_logic;

		-- Output port (non-WISHBONE signals):
		RXD             : in    std_logic;
		TXD             : out   std_logic;

		TXDONE          : out   std_logic;	-- 1pulse is Tx-Done
		TXEMPTY         : out   std_logic;	-- 0/1 = Tx not empty/Tx empty(Tx-Done)
		RXREADY         : out   std_logic;	-- 0/1 = Rx not ready/Rx ready
		RXFERR          : out   std_logic;	-- Rx Flame Error
		RXOERR          : out   std_logic	-- Rx OverRun Error
	);
end WBUART;

architecture RTL of WBUART is

	constant ZERO32         : std_logic_vector(31 downto 0) := (others=>'0');
	constant ONE32          : std_logic_vector(31 downto 0) := (others=>'1');

	constant DIVZERO        : std_logic_vector(7 downto 0) := (others=>'0');
	constant CNTZERO        : std_logic_vector(7 downto 0) := (others=>'0');
	constant CNTFULL        : std_logic_vector(7 downto 0) := (others=>'1');
	constant CNTTSET8       : std_logic_vector(7 downto 0) := "10110000";
	constant CNTRSET8       : std_logic_vector(7 downto 0) := "10011000";
	constant CNTTSET9       : std_logic_vector(7 downto 0) := "11000000";
	constant CNTRSET9       : std_logic_vector(7 downto 0) := "10101000";

	-- Address
	constant C_DATA         : std_logic_vector := x"00";
	constant C_STAT         : std_logic_vector := x"01";
	constant C_DIV_L        : std_logic_vector := x"02";
	constant C_DIV_H        : std_logic_vector := x"03";
	constant C_CTRL         : std_logic_vector := x"04";
	constant C_DATA9        : std_logic_vector := x"05";

	signal reg_ctrl         : std_logic_vector(4 downto 0);
	signal reg_divider      : std_logic_vector(11 downto 0);
	signal reg_data9        : std_logic;
	signal use_data9        : std_logic;
	signal rx_data9         : std_logic;
	signal divcnt           : std_logic_vector(11 downto 0);
	signal diven            : std_logic;
	signal wr_en            : std_logic;
	signal rd_en            : std_logic;
	signal rxd_is           : std_logic;
	signal txd_is           : std_logic;
	signal rxcnt            : std_logic_vector(7 downto 0);
	signal txcnt            : std_logic_vector(7 downto 0);
	signal rxsr             : std_logic_vector(8 downto 0);
	signal txsr             : std_logic_vector(9 downto 0);
	signal status           : std_logic_vector(7 downto 0);
	signal rxnext           : std_logic_vector(7 downto 0);
	signal txnext           : std_logic_vector(7 downto 0);
	signal rxoerr_is        : std_logic;
	signal rxferr_is        : std_logic;
	signal rxready_is       : std_logic;
	signal txempty_is       : std_logic;
	signal txdone_tmp       : std_logic_vector(1 downto 0);

begin

	wr_en <= '1' when WB_STB_I='1' and WB_WE_I='1' else '0';
	rd_en <= '1' when WB_STB_I='1' and WB_WE_I='0' else '0';

	-- Clock divider
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				divcnt <= (others=>'0');
				diven <= '0';
			elsif divcnt=DIVZERO then
				divcnt <= reg_divider;
				diven <= '1';
			else
				divcnt <= divcnt - '1';
				diven <= '0';
			end if;
		end if;
	end process;

	-- Receiver
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				rxd_is <= '1';
			else
				rxd_is <= RXD;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				rxcnt <= CNTFULL;
			elsif diven='1' then
				if rxcnt=CNTFULL then
					if rxd_is='0' then
						if use_data9='0' then
							rxcnt <= CNTRSET8;
						else
							rxcnt <= CNTRSET9;
						end if;
					else
						rxcnt <= CNTFULL;
					end if;
				else
					rxcnt <= rxcnt - 1;
				end if;
			else
				rxcnt <= rxcnt;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				rxsr <= (others=>'1');
			elsif diven='1' and rxcnt(3 downto 0)="0000" then
				rxsr <= rxd_is & rxsr(8 downto 1);
			end if;
		end if;
	end process;

	--
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' or (WB_ADR_I=C_DATA and rd_en='1') then
				rxready_is <= '0';
				rxferr_is  <= '0';
				rxoerr_is  <= '0';
			elsif diven='1' and rxcnt=CNTZERO then
				rxready_is <= '1';
				rxferr_is  <= not rxd_is;
				rxoerr_is  <= rxready_is;
			else
				rxready_is <= rxready_is;
				rxferr_is  <= rxferr_is;
				rxoerr_is  <= rxoerr_is;
			end if;
		end if;
	end process;

	--
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				rx_data9 <= '0';
				rxnext <= (others=>'0');
			elsif diven='1' and rxcnt=CNTZERO then
				rx_data9 <= rxsr(8);
				if use_data9='0' then
					rxnext <= rxsr(8 downto 1);
				else
					rxnext <= rxsr(7 downto 0);
				end if;
			else
				rx_data9 <= rx_data9;
				rxnext <= rxnext;
			end if;
		end if;
	end process;

	RXREADY <= rxready_is;
	RXFERR  <= rxferr_is;
	RXOERR  <= rxoerr_is;
	TXEMPTY <= txempty_is;

	-- Transmitter
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				txcnt <= CNTZERO;
			elsif txcnt=CNTZERO then
				if txempty_is='0' then
					if use_data9='0' then
						txcnt <= CNTTSET8;
					else
						txcnt <= CNTTSET9;
					end if;
				else
					txcnt <= CNTZERO;
				end if;
			elsif diven='1' then
				txcnt <= txcnt - 1;
			else
				txcnt <= txcnt;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				txsr <= (others=>'1');
			elsif txcnt=CNTZERO and txempty_is='0' then
				txsr <= reg_data9 & txnext & '0';
			elsif diven='1' and txcnt(3 downto 0)="0000" then
				txsr <= '1' & txsr(9 downto 1);
			else
				txsr <= txsr;
			end if;
		end if;
	end process;

	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				txd_is <= '1';
			elsif diven='1' and txcnt(3 downto 0)="0000" then
				txd_is <= txsr(0);
			else
				txd_is <= txd_is;
			end if;
		end if;
	end process;

	TXD <= txd_is;

	-- TX-Done
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				txdone_tmp <= (others=>'0');
			elsif txcnt=CNTZERO then
				txdone_tmp(0) <= '1';
			else
				txdone_tmp(0) <= '0';
			end if;

			txdone_tmp(1) <= txdone_tmp(0);

			if WB_RST_I='1' then
				TXDONE <= '0';
			elsif txdone_tmp="01" then
				TXDONE <= '1';
			else
				TXDONE <= '0';
			end if;
		end if;
	end process;

	--
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				txnext <= (others=>'0');
				txempty_is <= '1';
			elsif WB_ADR_I=C_DATA and wr_en='1' then
				txnext <= WB_DAT_I;
				txempty_is <= '0';
			elsif txcnt=CNTZERO then
				txnext <= txnext;
				txempty_is <= '1';
			else
				txnext <= txnext;
				txempty_is <= txempty_is;
			end if;
		end if;
	end process;

	-- Register:DIVIDER
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				reg_divider <= DIVIDER_INIT;
			elsif WB_ADR_I=C_DIV_L and wr_en='1' then
				reg_divider(WB_DAT_I'range) <= WB_DAT_I;
			elsif WB_ADR_I=C_DIV_H and wr_en='1' then
				reg_divider(11 downto 8) <= WB_DAT_I(3 downto 0);
			else
				reg_divider <= reg_divider;
			end if;
		end if;
	end process;

	-- Register:CTRL
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				reg_ctrl <= (others=>'0');
			elsif WB_ADR_I=C_CTRL and wr_en='1' then
				reg_ctrl <= WB_DAT_I(reg_ctrl'range);
			else
				reg_ctrl <= reg_ctrl;
			end if;
		end if;
	end process;

	use_data9 <= reg_ctrl(4);

	-- Register:DATA9
	process (WB_CLK_I)
	begin
		if WB_CLK_I'event and WB_CLK_I='1' then
			if WB_RST_I='1' then
				reg_data9 <= '1';
			elsif WB_ADR_I=C_DATA9 and wr_en='1' then
				reg_data9 <= WB_DAT_I(0);
			else
				reg_data9 <= reg_data9;
			end if;
		end if;
	end process;

	--
	WB_ACK_O <= WB_STB_I;

	--
	status <= "0000" & rxoerr_is & rxferr_is & rxready_is & txempty_is;

	--
	WB_DAT_O <= rxnext when WB_ADR_I=C_DATA else
		    status when WB_ADR_I=C_STAT else
		    reg_divider(WB_ADR_I'range) when WB_ADR_I=C_DIV_L else
		    ZERO32(7 downto 4) & reg_divider(11 downto 8) when WB_ADR_I=C_DIV_H else
		    ZERO32(WB_DAT_O'left downto reg_ctrl'left+1) & reg_ctrl when WB_ADR_I=C_CTRL else
		    ZERO32(WB_DAT_O'left downto 1) & rx_data9 when WB_ADR_I=C_DATA9 else
		    (others=>'0');

	--
	WB_IRQ_O <= (rxoerr_is and reg_ctrl(3)) or
		    (rxferr_is and reg_ctrl(2)) or
		    (rxready_is and reg_ctrl(1)) or
		    (txempty_is and reg_ctrl(0));

end RTL;
